Package: ScaledMatrix
Version: 1.20.0
Date: 2024-02-29
Title: Creating a DelayedMatrix of Scaled and Centered Values
Authors@R: person("Aaron", "Lun", role=c("aut", "cre", "cph"),
        email="infinite.monkeys.with.keyboards@gmail.com")
Imports: methods, Matrix, S4Vectors, DelayedArray
Suggests: testthat, BiocStyle, knitr, rmarkdown, BiocSingular,
        DelayedMatrixStats
biocViews: Software, DataRepresentation
Description: 
    Provides delayed computation of a matrix of scaled and centered values.
    The result is equivalent to using the scale() function but avoids explicit 
    realization of a dense matrix during block processing. This permits greater
    efficiency in common operations, most notably matrix multiplication.
License: GPL-3
VignetteBuilder: knitr
RoxygenNote: 7.3.1
BugReports: https://github.com/LTLA/ScaledMatrix/issues
URL: https://github.com/LTLA/ScaledMatrix
git_url: https://git.bioconductor.org/packages/ScaledMatrix
git_branch: RELEASE_3_23
git_last_commit: 94df0ea
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:23:30 UTC; biocbuild
Author: Aaron Lun [aut, cre, cph]
Maintainer: Aaron Lun <infinite.monkeys.with.keyboards@gmail.com>
Built: R 4.6.0; ; 2026-04-29 04:59:08 UTC; unix
