# Centered and scaled matrix

|Build|Status|
|-----|----|
| [Bioc-release](https://bioconductor.org/packages/release/bioc/html/ScaledMatrix.html) | [![](https://bioconductor.org/shields/build/release/bioc/ScaledMatrix.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/ScaledMatrix) |
| [Bioc-devel](https://bioconductor.org/packages/devel/bioc/html/ScaledMatrix.html) | [![](https://bioconductor.org/shields/build/devel/bioc/ScaledMatrix.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/ScaledMatrix) | 

This package performs the delayed centering and scaling of a matrix, e.g., before singular value decomposition for a principal components analysis.
It implements a number of efficient methods for operating on the delayed matrix of residuals, most notably matrix multiplication and calculation of row/column sums or means.
Check out the user's guide on the [Bioconductor landing page](https://bioconductor.org/packages/devel/bioc/html/ScaledMatrix.html) for more details.
